#####################################################################
#
#	Routines for bivariate GEV model fitting and inference.
#	Requires the ismev library (available from CRAN).
#
#####################################################################
require(ismev)
#####################################################################
# Function to calculate log density of a GEV distribution as a
# vector. Arguments:
#
# theta Parameter vector
# y	Vector of observed maxima
# X	Design matrix
# mul   Selects columns of X for modelling location parameter
# sigl  Ditto, dispersion parameter
# shl   Ditto, shape parameter
#
# Value: a vector, the same length as y, with each element containing 
# the log density for a GEV distribution evaluated at the corresponding
# y value. 
# 
#####################################################################
gev.logdens <- function(theta,y,X=NULL,mul=NULL,sigl=NULL,shl=NULL) {
 n <- length(y)
 nmu <- 1
 nsig <- 1
 nxi <- 1
 if(!is.null(mul)) nmu <- nmu + length(mul)
 if(!is.null(sigl)) nsig <- nsig + length(sigl)
 if(!is.null(shl)) nxi <- nxi + length(shl)
 mupar <- theta[1:nmu]
 sigpar <- theta[(nmu+1):(nmu+nsig)]
 xipar <- theta[(nmu+nsig+1):(nmu+nsig+nxi)]
 ones <- rep(1,n)
 if (is.null(mul)) {
  mu <- rep(mupar,n)
 } else {
  mu <- cbind(ones,X[,mul]) %*% mupar
 }
 if (is.null(sigl)) {
  sigma <- rep(sigpar,n)
 } else {
  sigma <- cbind(ones,X[,sigl]) %*% sigpar
 }
 if (is.null(shl)) {
  xi <- rep(xipar,n)
 } else {
  xi <- cbind(ones,X[,shl]) %*% xipar
 }
 gum <- (xi == 0)
 z <- rep(NA,n)
 yy <- (y - mu) / sigma
 z[gum] <-  -yy[gum] - exp(-yy[gum]) - log(sigma[gum])
 yy <- 1 + xi[!gum]*yy[!gum]
 z[!gum] <- - yy^(-1/xi[!gum]) - ((1+1/xi[!gum])*log(yy)) - log(sigma[!gum])
 z[yy < 0] <- -Inf
 z
}
######################################################################
# Negative log-likelihood of GEV distribution (alternative to 
# gev.lik). Arguments:
#
# theta 	Parameter vector over which (potentially) likelihood
#		is to be maximised
# y		Vector of observed maxima
# X		Design matrix
# mul   	Selects columns of X for modelling location parameter
# sigl  	Ditto, dispersion parameter
# shl   	Ditto, shape parameter
# theta.full	Full parameter vector
# which.fixed	If theta is shorter than theta.full, a vector identifying
#		the positions in theta.full that should be treated as 
#		fixed
# dep.adj	Carry out dependence adjustment?
# thetahat	If dep.adj, the MLE of theta
# C		If dep.adj, the matrix of the transformation
######################################################################
gev.negloglik <- function(theta,y,X=NULL,mul=NULL,sigl=NULL,shl=NULL,
			theta.full=theta,which.fixed=NULL,
			dep.adj=FALSE,thetahat,C) {
 if (is.null(which.fixed)) {
  z <- theta
 } else {
  z <- theta.full
  z[-which.fixed] <- theta
 }
 if (!dep.adj) {
 thetastar <- z
 } else {
  thetastar <- thetahat + (C %*% (z - thetahat))
 }
 z <- -sum(gev.logdens(thetastar,y,X,mul,sigl,shl))
 z
}
######################################################################
# To fit a GEV distribution (alternative to gev.fit, allowing 
# profiling over any parameters). Arguments as gev.negloglik.
# Should be called with a starting value for theta. If theta is 
# a subset of theta.full, the other parameters (identified by
# which.fixed) are fixed.
######################################################################
gev.fit2 <- function(theta,y,X=NULL,mul=NULL,sigl=NULL,shl=NULL,
 	             theta.full=theta,which.fixed=NULL,
			dep.adj=FALSE,thetahat,C) {
 init <- gev.negloglik(theta,y=y,X=X,mul=mul,sigl=sigl,shl=shl,
                        theta.full=theta.full,which.fixed=which.fixed)
 z <- nlm(gev.negloglik,theta,hessian=TRUE,fscale=init,
		y=y,X=X,mul=mul,sigl=sigl,shl=shl,
                theta.full=theta.full,which.fixed=which.fixed,
		dep.adj=dep.adj,thetahat=thetahat,C=C)
 H <- (z$hessian + t(z$hessian)) / 2
 mle <- z$estimate
 Hinv <- solve(H)
 fit <- list(mle=mle,nllh=z$minimum,conv=z$code,se=diag(sqrt(Hinv)),cov=Hinv)
 fit
}
######################################################################
# To calculate a profile likelihood over 1 parameter in a GEV 
# model. Arguments:
#
# model		A fitted GEV model (either from gev.fit or gev.fit2)
# prof.par	The number of the parameter over which to compute
#		the profile
# lower		The lower bound for the parameter
# upper		The upper bound for the parameter
# npts		The number of points at which the profile will be
#		computed
# 
# All other arguments as for gev.fit2
#
# Value: a matrix ((npts+1)*2) containing the parameter values and 
#        profile log-likelihoods
######################################################################
gev.prof1d <- function(model,prof.par,lower=NULL,upper=NULL,npts=20,
			y,X=NULL,mul=NULL,sigl=NULL,shl=NULL,
 	             	dep.adj=FALSE,C) {
 warnopt <- options()$warn; options(warn=-1)
 if (is.null(lower)) lower <- model$mle[prof.par] - 3*model$se[prof.par]
 if (is.null(upper)) upper <- model$mle[prof.par] + 3*model$se[prof.par]
 par.grid <- sort(c(seq(lower,upper,length=npts),model$mle[prof.par]))
 z <- rep(NA,npts+1)
 theta <- model$mle[-prof.par]
 theta.full <- model$mle
#
# To avoid illegal starting values, move from the MLE downwards and
# then back up, at each stage using the previous estimate as the
# starting value
#
 mle.idx <- match(model$mle[prof.par],par.grid)
 eval.order <- c(rev(1:(mle.idx-1)),mle.idx:(npts+1))
 for (i in eval.order) {
  theta.full[prof.par] <- par.grid[i]
  zz <- try(gev.fit2(theta,y=y,X=X,mul=mul,sigl=sigl,shl=shl,
 	            theta.full=theta.full,which.fixed=prof.par,
	            dep.adj=dep.adj,thetahat=model$mle,C=C),silent=TRUE)
  if (class(zz) == "try-error") {
   z[i] <- NA
  } else {
   z[i] <- -zz$nllh
   if (i > 1) { 
    theta <- zz$mle
   } else {
    theta <- model$mle[-prof.par]
   }
  }
 }
 options(warn=warnopt)
 keep <- !duplicated(par.grid)
 par.grid <- par.grid[keep]
 z <- z[keep]
 data.frame(parval=par.grid,Logl=z)
}
######################################################################
# To calculate a profile likelihood over 2 parameters in a GEV 
# model. Arguments:
#
# model		A fitted GEV model (either from gev.fit or gev.fit2)
# prof.par	A vector of 2 integers indicating the numbers of the 
#		parameters over which to compute the profile
# lower		The lower bounds for the parameters
# upper		The upper bounds for the parameters
# npts		The numbers of points, in each direction, at which 
#               the profile will be computed
# 
# All other arguments as for gev.fit2
#
# Value: a list of three elements:
#
# grid1		The values of the first parameter in prof.par
# grid2		The values of the second parameter in prof.par
# Logl		A matrix of profile log-likelihoods
######################################################################
gev.prof2d <- function(model,prof.par,lower=NULL,upper=NULL,npts=c(10,10),
			y,X=NULL,mul=NULL,sigl=NULL,shl=NULL,
 	             	dep.adj=FALSE,C) {
 if (length(prof.par) !=2) stop("prof.par must have 2 elements")
 warnopt <- options()$warn; options(warn=-1)
 if (is.null(lower)) lower <- model$mle[prof.par] - 3*model$se[prof.par]
 if (is.null(upper)) upper <- model$mle[prof.par] + 3*model$se[prof.par]
 grid1 <- sort(c(seq(lower[1],upper[1],length=npts[1]),model$mle[prof.par[1]]))
 grid2 <- sort(c(seq(lower[2],upper[2],length=npts[2]),model$mle[prof.par[2]]))
 z <- matrix(nrow=npts[1]+1,ncol=npts[2]+1)
 theta.start <- model$mle[-prof.par]
 theta.full <- model$mle
#
# To avoid illegal starting values, move from the MLE downwards and
# then back up, at each stage using neighbouring estimates as 
# starting values.
#
 mle.idx1 <- match(model$mle[prof.par[1]],grid1)
 mle.idx2 <- match(model$mle[prof.par[2]],grid2)
 eval.order1 <- c(rev(1:(mle.idx1-1)),mle.idx1:(npts[1]+1))
 eval.order2 <- c(rev(1:(mle.idx2-1)),mle.idx2:(npts[2]+1))
 start.array <- array(dim=c(length(theta.start),length(grid1),length(grid2)))
 start.array[,mle.idx1,mle.idx2] <- theta.start
 for (i in eval.order1) {
  i.nbr <- max(i-1,1):(i+1)
  i.nbr <- i.nbr[i.nbr <= length(grid1)]
  for (j in eval.order2) {
   j.nbr <- max(j-1,1):(j+1)
   j.nbr <- j.nbr[j.nbr <= length(grid2)]
   theta <- rowMeans(start.array[,i.nbr,j.nbr],na.rm=TRUE)
   if (any(is.na(theta))) theta <- theta.start
   theta.full[prof.par] <- c(grid1[i],grid2[j])
   zz <- try(gev.fit2(theta,y=y,X=X,mul=mul,sigl=sigl,shl=shl,
  	            theta.full=theta.full,which.fixed=prof.par,
 	            dep.adj=dep.adj,thetahat=model$mle,C=C),silent=TRUE)
   if (class(zz) == "try-error") {
    z[i,j] <- NA
   } else {
    z[i,j] <- -zz$nllh
    start.array[,i,j] <- zz$mle
    if (j == mle.idx2) theta.start <- theta
   }
  }
 }
 keeprow <- !duplicated(grid1)
 keepcol <- !duplicated(grid2)
 grid1 <- grid1[keeprow]; grid2 <- grid2[keepcol]
 z <- z[keeprow,keepcol]
 options(warn=warnopt)
 list(grid1=grid1,grid2=grid2,Logl=z)
}
######################################################################
# To compute the sandwich estimate of the covariance matrix for
# a GEV fitted to clustered data. Arguments:
#
# mle		A vector of GEV parameter estimates
# y             The data from which mle was derived
# X		The design matrix used to obtain model
# mul etc.	The parameterisation used to obtain model
# clusters	A vector, the same length as model$data, indicating
#		cluster membership
#
# Value: a list of three matrices: 
# 
# cov 	the sandwich covariance matrix estimate
# H	the Hessian of the independence log-likelihood
# V	the estimated covariance matrix of the score vector
######################################################################
gev.sandwich <- function(mle,y,X=NULL,mul=NULL,sigl=NULL,shl=NULL,
				clusters=1:length(y)) {
 H <- num.hess(gev.negloglik,mle,y=y,
				X=X,mul=mul,sigl=sigl,shl=shl)$hessian
 np <- length(mle)
 V <- matrix(0,nrow=np,ncol=np)
 for (k in unique(clusters)) {
  wanted <- clusters == k
  if (is.null(X)) {
   Xrows <- matrix(rep(1,sum(wanted)),ncol=1)
  } else {
   Xrows <- matrix(X[wanted,],ncol=ncol(X))
  }
  U <- num.deriv(gev.negloglik,mle,y=y[wanted],X=Xrows,
                                   mul=mul,sigl=sigl,shl=shl)$gradient
  V <- V + (U %*% t(U))
 }
 Hinv <- solve(H)
 list(cov=Hinv %*% V %*% Hinv,H=H,V=V)
}
#####################################################################
# Function to calculate log density of a bivariate logistic 
# distribution with GEV margins. Arguments as gev.logdens,
# except that the responses and design matrix are split into
# 2 (y1 & y2, X1 & X2) so that the pairs can be identified.
# Also, the last element of theta is alpha (the dependence
# parameter)
#####################################################################
gev2.logdens <- function(theta,y1,y2,X1=NULL,X2=NULL,mul=NULL,
                                           sigl=NULL,shl=NULL) {
 n <- length(y1)
 nmu <- 1
 nsig <- 1
 nxi <- 1
 if(!is.null(mul)) nmu <- nmu + length(mul)
 if(!is.null(sigl)) nsig <- nsig + length(sigl)
 if(!is.null(shl)) nxi <- nxi + length(shl)
 mupar <- theta[1:nmu]
 sigpar <- theta[(nmu+1):(nmu+nsig)]
 xipar <- theta[(nmu+nsig+1):(nmu+nsig+nxi)]
 alpha <- theta[nmu+nsig+nxi+1]
 ones <- rep(1,n)
 if (is.null(mul)) {
  mu1 <- rep(mupar,n)
  mu2 <- rep(mupar,n)
 } else {
  mu1 <- cbind(ones,X1[,mul]) %*% mupar
  mu2 <- cbind(ones,X2[,mul]) %*% mupar
 }
 if (is.null(sigl)) {
  sigma1 <- rep(sigpar,n)
  sigma2 <- rep(sigpar,n)
 } else {
  sigma1 <- cbind(ones,X1[,sigl]) %*% sigpar
  sigma2 <- cbind(ones,X2[,sigl]) %*% sigpar
 }
 if (is.null(shl)) {
  xi1 <- rep(xipar,n)
  xi2 <- rep(xipar,n)
 } else {
  xi1 <- cbind(ones,X1[,shl]) %*% xipar
  xi2 <- cbind(ones,X2[,shl]) %*% xipar
 }
 z1 <- (1 + (xi1*(y1-mu1)/sigma1))
 z2 <- (1 + (xi2*(y2-mu2)/sigma2))
 zz1 <- z1^(1/xi1); zz2 <- z2^(1/xi2)
 logG <- -(zz1^(-1/alpha)+zz2^(-1/alpha))^alpha
#
# Distribution function is for transformed variables zz1 & zz2 -
# need joint density for y1 & y2, so don't forget the Jacobian
# (last 2 lines below)
#
 logdens <- logG + log(-logG - 1 + 1/alpha) +
            ((alpha-2)*log(-logG)/alpha) - 
            ((alpha+1) * (log(zz1)+log(zz2)) / alpha) +
            (((1/xi1) - 1) * log(z1)) + (((1/xi2) - 1) * log(z2)) -
            log(sigma1) - log(sigma2)
 logdens
}
######################################################################
# Negative log-likelihood of bivariate GEV distribution. Arguments 
# in the same relation to those of gev2.logdens as are those of
# gev.negloglik with respect to gev.logdens.
######################################################################
gev2.negloglik <- function(theta,y1,y2,X1=NULL,X2=NULL,
                           mul=NULL,sigl=NULL,shl=NULL,
 			   theta.full=theta,which.fixed=NULL) {
 if (is.null(which.fixed)) {
  z <- theta
 } else {
  z <- theta.full
  z[-which.fixed] <- theta
 }
 -sum(gev2.logdens(z,y1=y1,y2=y2,X1=X1,X2=X2,mul=mul,sigl=sigl,shl=shl))
}
######################################################################
# To fit a bivariate GEV distribution. Arguments as gev2.negloglik.
# Should be called with a starting value for theta. If theta is 
# a subset of theta.full, the other parameters (identified by
# which.fixed) are fixed.
######################################################################
gev2.fit <- function(theta,y1,y2,X1=NULL,X2=NULL,
                     mul=NULL,sigl=NULL,shl=NULL,
 	             theta.full=theta,which.fixed=NULL) {
 init <- gev2.negloglik(theta,y1=y1,y2=y2,X1=X1,X2=X2,
			mul=mul,sigl=sigl,shl=shl,
                        theta.full=theta.full,which.fixed=which.fixed)
 z <- nlm(gev2.negloglik,theta,hessian=TRUE,fscale=init,
		y1=y1,y2=y2,X1=X1,X2=X2,mul=mul,sigl=sigl,shl=shl,
                theta.full=theta.full,which.fixed=which.fixed)
 H <- (z$hessian + t(z$hessian)) / 2
 mle <- z$estimate
 Hinv <- solve(H)
 fit <- list(mle=mle,nllh=z$minimum,conv=z$code,se=diag(sqrt(Hinv)),cov=Hinv)
 fit
}
######################################################################
# To calculate a profile likelihood over 1 parameter in a bivariate GEV 
# model. Arguments:
#
# model		A fitted model (from gev2.fit)
# prof.par	The number of the parameter over which to compute
#		the profile
# lower		The lower bound for the parameter
# upper		The upper bound for the parameter
# npts		The number of points at which the profile will be
#		computed
# 
# All other arguments as for gev.fit2
#
# Value: a matrix ((npts+1)*2) containing the parameter values and 
#        profile log-likelihoods
######################################################################
gev2.prof1d <- function(model,prof.par,lower=NULL,upper=NULL,npts=20,
			y1,y2,X1=NULL,X2=NULL,mul=NULL,sigl=NULL,shl=NULL) {
 warnopt <- options()$warn; options(warn=-1)
 if (is.null(lower)) lower <- model$mle[prof.par] - 3*model$se[prof.par]
 if (is.null(upper)) upper <- model$mle[prof.par] + 3*model$se[prof.par]
 par.grid <- sort(c(seq(lower,upper,length=npts),model$mle[prof.par]))
 z <- rep(NA,npts)
 theta <- model$mle[-prof.par]
 theta.full <- model$mle
#
# To avoid illegal starting values, move from the MLE downwards and
# then back up, at each stage using the previous estimate as the
# starting value
#
 mle.idx <- match(model$mle[prof.par],par.grid)
 eval.order <- c(rev(1:(mle.idx-1)),mle.idx:(npts+1))
 for (i in eval.order) {
  theta.full[prof.par] <- par.grid[i]
  zz <- try(gev2.fit(theta,y1=y1,y2=y2,X1=X1,X2=X2,mul=mul,sigl=sigl,shl=shl,
 	            theta.full=theta.full,which.fixed=prof.par),silent=TRUE)
  if (class(zz) == "try-error") {
   z[i] <- NA
  } else {
   z[i] <- -zz$nllh
   if (i > 1) { 
    theta <- zz$mle
   } else {
    theta <- model$mle[-prof.par]
   }
  }
 }
 options(warn=warnopt)
 keep <- !duplicated(par.grid)
 par.grid <- par.grid[keep]
 z <- z[keep]
 data.frame(parval=par.grid,Logl=z)
}
######################################################################
# To calculate a profile likelihood over 2 parameters in a bivariate GEV 
# model. Arguments:
#
# model		A fitted GEV model (either from gev.fit or gev.fit2)
# prof.par	A vector of 2 integers indicating the numbers of the 
#		parameters over which to compute the profile
# lower		The lower bounds for the parameters
# upper		The upper bounds for the parameters
# npts		The numbers of points, in each direction, at which 
#               the profile will be computed
# 
# All other arguments as for gev2.fit
#
# Value: a list of three elements:
#
# grid1		The values of the first parameter in prof.par
# grid2		The values of the second parameter in prof.par
# Logl		A matrix of profile log-likelihoods
######################################################################
gev2.prof2d <- function(model,prof.par,lower=NULL,upper=NULL,
			npts=c(10,10),y1=y1,y2=y2,X1=NULL,X2=NULL,
			mul=NULL,sigl=NULL,shl=NULL) {
 if (length(prof.par) !=2) stop("prof.par must have 2 elements")
 warnopt <- options()$warn; options(warn=-1)
 if (is.null(lower)) lower <- model$mle[prof.par] - 3*model$se[prof.par]
 if (is.null(upper)) upper <- model$mle[prof.par] + 3*model$se[prof.par]
 grid1 <- sort(c(seq(lower[1],upper[1],length=npts[1]),model$mle[prof.par[1]]))
 grid2 <- sort(c(seq(lower[2],upper[2],length=npts[2]),model$mle[prof.par[2]]))
 z <- matrix(nrow=npts[1]+1,ncol=npts[2]+1)
 theta.start <- model$mle[-prof.par]
 theta.full <- model$mle
#
# To avoid illegal starting values, move from the MLE downwards and
# then back up, at each stage using the previous estimate as the
# starting value
#
 mle.idx1 <- match(model$mle[prof.par[1]],grid1)
 mle.idx2 <- match(model$mle[prof.par[2]],grid2)
 eval.order1 <- c(rev(1:(mle.idx1-1)),mle.idx1:(npts[1]+1))
 eval.order2 <- c(rev(1:(mle.idx2-1)),mle.idx2:(npts[2]+1))
 start.array <- array(dim=c(length(theta.start),length(grid1),length(grid2)))
 start.array[,mle.idx1,mle.idx2] <- theta.start
 for (i in eval.order1) {
  i.nbr <- max(i-1,1):(i+1)
  i.nbr <- i.nbr[i.nbr <= length(grid1)]
  for (j in eval.order2) {
   j.nbr <- max(j-1,1):(j+1)
   j.nbr <- j.nbr[j.nbr <= length(grid2)]
   theta <- rowMeans(start.array[,i.nbr,j.nbr],na.rm=TRUE)
   if (any(is.na(theta))) theta <- theta.start
   theta.full[prof.par] <- c(grid1[i],grid2[j])
   zz <- try(gev2.fit(theta,y1=y1,y2=y2,X1=X1,X2=X2,mul=mul,sigl=sigl,
                      shl=shl,theta.full=theta.full,
                      which.fixed=prof.par),silent=TRUE)
   if (class(zz) == "try-error") {
    z[i,j] <- NA
   } else {
    z[i,j] <- -zz$nllh
    start.array[,i,j] <- zz$mle
    if (j == mle.idx2) theta.start <- theta
   }
  }
 }
 keeprow <- !duplicated(grid1)
 keepcol <- !duplicated(grid2)
 grid1 <- grid1[keeprow]; grid2 <- grid2[keepcol]
 z <- z[keeprow,keepcol]
 options(warn=warnopt)
 list(grid1=grid1,grid2=grid2,Logl=z)
}
######################################################################
# To compute the transformation matrix needed for the dependence-adjusted
# likelihood. Arguments:
#
# Hind		Hessian of "independence" log-likelihood
# H		Hessian of adjusted inference function
# method	Method to use for calculating matrix square root -
#		"cholesky" or "dilation"
#
# Value: the required transformation matrix
######################################################################
adj.trans <- function(Hind,H,method) {
 if (method == "cholesky") {
  Mind <- chol(Hind)
  M <- chol(H)
 } else if (method == "dilation") {
  z <- eigen(Hind,symmetric=TRUE)
  Mind <- z$vectors %*% diag(sqrt(z$values)) %*% t(z$vectors)
  z <- eigen(H,symmetric=TRUE)
  M <- z$vectors %*% diag(sqrt(z$values)) %*% t(z$vectors)
 } else {
  stop("Allowed methods are cholesky and dilation")
 }
 solve(Mind) %*% M
}
######################################################################
#       Numerical differentiation (because there aren't any 
#	user-friendly routines for this in R). Arguments:
#
#	f	The function to be differentiated
#	x0	The point at which to evaluate the gradient
#	lb	Vector of lower bounds on elements of x (either
#		a scalar, or a vector the same length as x0)
#	ub	Vector of upper bounds
#	steps	Vector of largest step sizes to consider when
#		evaluating the gradient. Defaults to the larger
#		of abs(x0)/10^4 and 0.001.
#	shrink	Amount by which to shrink the stepsize at each
#		iteration, to obtain an improved estimate
#       maxit   Maximum number of iterations
#       relerr  Relative error required
#	...	Any other arguments to f
#
#       This uses Ridders' method - see Press et al. (1992), Section
#	5.7 for example. The implementation here is basically that
#	given by Press et al., translated from FORTRAN to R and
#       adapted to handle boundaries (in which case a right- or 
#	left-hand derivative is evaluated as appropriate) and 
#	vector arguments. 1-sided derivatives are calculated by 
#       reflecting the function about the boundary, which is
#	equivalent to using function evaluations on one side of the
#	boundary only - and correspondingly less accurate. For
#	1-sided derivatives it may therefore be necessary to 
#	decrease the default value of steps to obtain the required 
#	accuracy.
#
#	The value of the function is a list containing two elements,
#	as follows:
#
#	gradient	vector of length length(x0), containing
#			the calculated derivatives
#	error		Estimate of the errors in the computed
#			gradients
#	
######################################################################
num.deriv <- function(f,x0,xlo=-Inf,xhi=Inf,steps=NULL,shrink=1.4,
					maxit=10,relerr=1e-6,...) {
 np <- length(x0); d <- numeric(np)
 nb <- length(xlo)
 if (nb > 1 & nb != np) stop("xlo and x0 must be the same length")
 nb <- length(xhi)
 if (nb > 1 & nb != np) stop("xhi and x0 must be the same length")
 if (is.null(steps)) {
  h <- pmax(abs(x0)*1e-4,1e-3) 
 } else {
  h <- steps
 }
 right <- (x0 == xlo); left <- (x0 == xhi)
 f0 <- f(x0,...)
 shr2 <- shrink^2; err <- rep(Inf,np)
 delta <- .Machine$double.eps
#
# Set initial step sizes for each parameter. 
#
 hh <- pmin(h,(x0-xlo-delta),(xhi-delta-x0))
 hh[left] <- pmin(h[left],(xhi-xlo-delta)[left])
 hh[right] <- pmin(h[right],(xhi-xlo-delta)[right])
 tight <- right & left; d[tight] <- NA
 wanted.els <- (1:np)[!tight]
 for (p in wanted.els) {
  a <- matrix(nrow=maxit,ncol=maxit)
  i <- 1; cnvrge <- FALSE
  while (i <= maxit & !cnvrge) {
   xx <- x0
   if (!right[p]) {
    xlo <- x0[p] - hh[p]; xx[p] <- xlo; f1 <- f(xx,...)
#
#	For left derivatives, extrapolate symmetrically beyond the
#	upper limit
#
    if (left[p]) f2 <- (2*f0) - f1
   }
   if (!left[p]) {
    xhi <- x0[p] + hh[p]; xx[p] <- xhi; f2 <- f(xx,...)
#
#	And for right derivatives, extrapolate beyond the lower
#	limit
#
    if (right[p]) f1 <- (2*f0) - f2
   }
   a[1,i] <- (f2 - f1) / (2*hh[p])
#
#	This is the j-loop in Numerical Recipes. Although looping
#	is normally to be avoided in R, experimentation suggests
#	that in this case it's actally quicker than the corresponding
#	vector-based calculation - partly do do with the fact that 
#	in that case you have do do things like identify the 
#	location of the smallest value in the column etc., I guess
#
   if (i > 1) {
    fac <- shr2
    for (j in 2:i) {
     a[j,i] <- (a[j-1,i]*fac-a[j-1,i-1]) / (fac-1)
     fac <- shr2*fac
     errt <- max(abs(a[j,i]-a[j-1,i]),abs(a[j,i]-a[j-1,i-1]))
     if (is.na(errt)) errt <- Inf
     if (errt < err[p]) { 
      err[p] <- errt
      d[p] <- a[j,i]
     }
    }
     cnvrge <- (abs(a[i,i]-a[i-1,i-1]) > 2*err[p])
     if (is.na(cnvrge)) cnvrge <- FALSE
     if (abs(d[p]) > relerr) {
      cnvrge <- cnvrge & (abs(err[p]) < abs(relerr * d[p]))
     }
    }
   hh[p] <- hh[p] / shrink; i <- i+1
  }
 }
 list(gradient=d,error=err)
}
######################################################################
#       Numerical evaluation of a Hessian. Arguments:
#
#	f	The function to be differentiated
#	x0	The point at which to evaluate the Hessian
#	xlo	Vector of lower bounds on elements of x (either
#		a scalar, or a vector the same length as x0)
#	xhi	Vector of upper bounds
#	steps	Vector of largest step sizes to consider for 
#		finite-difference approximations. Default is
#               approximately the square root of than that in 
#		num.deriv, because the denominator in each 
#		approximation is steps^2)
#	shrink	Amount by which to shrink the stepsize at each
#		iteration, to obtain an improved estimate
#       maxit   Maximum number of iterations
#       relerr  Relative error required
#	...	Any other arguments to f
#
#       This uses polynomial extrapolation to the limit, in exactly
#       the same way as num.deriv(). The result is regarded is the
#       limit of a sequence of finite-difference approximations, 
#       evaluated at a sequence of points e1, e2, e3, ... 
#       which represent progressively smaller multiples of the 
#       initial value of steps. The finite difference approximation 
#	actually used is, for off-diagonal elements,
#
#	{f(x1+h1,x2+h2)+f(x1-h1,x2-h2)-[f(x1+h1,x2-h2)+f(x1-h1,x2+h2)]}
#       _______________________________________________________________
#				4*h1*h2
#
#	or (f1 + f2 - f3 - f4) / (4*h1*h2), where h1 and h2 are 
#	multiples of the e's. For diagonal elements the approximation
#	is
#
#	{f(x+h) - f(x) - f(x) + f(x-h)} / h
#
#	- NB the use of h rather than 4h is *correct* here - repeated
#       application of the first-order finite difference calculation
#       yields a step size of 2h, which has been halved in this case.
#
#	The value of the function is a list containing two elements,
#	as follows:
#
#	hessian		matrix containing the calculated Hessian
#	error		Estimate of the errors in the computed
#			hessian
#	
######################################################################
num.hess <- function(f,x0,xlo=-Inf,xhi=Inf,steps=NULL,shrink=1.4,
					maxit=10,relerr=1e-6,...) {
 np <- length(x0); d <- matrix(rep(0,np*np),nrow=np,ncol=np)
 nb <- length(xlo)
 if (nb > 1 & nb != np) stop("xlo and x0 must be the same length")
 nb <- length(xhi)
 if (nb > 1 & nb != np) stop("xhi and x0 must be the same length")
 if (is.null(steps)) {
  h <- pmax(abs(x0)*1e-2,1e-2) 
 } else {
  h <- steps
 }
 right <- (x0 == xlo); left <- (x0 == xhi)
 tight <- right & left
 if (any((right | left)[!tight])) {
   stop("I don't know how to evaluate Hessian on a boundary")
 }
 shr2 <- shrink^2; err <- matrix(rep(Inf,np*np),nrow=np)
 delta <- .Machine$double.eps
 oppsign <- c(1,-1)
 f0 <- f(x0,...)
 for (p1 in (1:np)[!tight]) {
  for (p2 in (1:np)[!tight]) {
#
# Reset initial step sizes for each parameter. No need to set 
# tight values to NA here, because d contains NA on initialisation
#
   hh <- pmin(h,(x0-xlo-delta),(xhi-delta-x0))
   a <- matrix(nrow=maxit,ncol=maxit); els <- c(p1,p2)
   i <- 1; cnvrge <- FALSE
   while (i <= maxit & !cnvrge) {
    xx <- x0
#
# Different finite difference approximations required on and off
# the diagonal
#
    if (p1 == p2) {
     xnew <- x0[p1] - hh[p1]; xx[p1] <- xnew; f1 <- f(xx,...)
     xnew <- x0[p1] + hh[p1]; xx[p1] <- xnew; f2 <- f(xx,...)
     a[1,i] <- ((f1 - f0) - (f0 - f2)) / (hh[p1]*hh[p1])
    } else {
     xnew <- x0[els] + hh[els]; xx[els] <- xnew; f1 <- f(xx,...)
     xnew <- x0[els] - hh[els]; xx[els] <- xnew; f2 <- f(xx,...)
     xnew <- x0[els] + (hh[els]*oppsign); xx[els] <- xnew; f3 <- f(xx,...)
     xnew <- x0[els] - (hh[els]*oppsign); xx[els] <- xnew; f4 <- f(xx,...)
     a[1,i] <- (f1 + f2 - f3 - f4) / (4*hh[p1]*hh[p2])
   }
#
#	This is the j-loop in Numerical Recipes. Although looping
#	is normally to be avoided in R, experimentation suggests
#	that in this case it's actally quicker than the corresponding
#	vector-based calculation - partly do do with the fact that 
#	in that case you have do do things like identify the 
#	location of the smallest value in the column etc., I guess
#
    if (i > 1) {
     fac <- shr2 
     for (j in 2:i) {
      a[j,i] <- (a[j-1,i]*fac-a[j-1,i-1]) / (fac-1)
      fac <- shr2*fac
      errt <- max(abs(a[j,i]-a[j-1,i]),abs(a[j,i]-a[j-1,i-1]))
      if (is.na(errt)) errt <- Inf
      if (errt < err[p1,p2]) { 
       err[p1,p2] <- errt
       d[p1,p2] <- a[j,i]
      }
     }
     cnvrge <- (abs(a[i,i]-a[i-1,i-1]) > 2*err[p1,p2]) 
     if (is.na(cnvrge)) cnvrge <- FALSE
     if (abs(d[p1,p2]) > relerr) {
      cnvrge <- cnvrge & (abs(err[p1,p2]) < abs(relerr * d[p1,p2]))
     }
    }
    hh[els] <- hh[els] / shrink; i <- i+1
   }
  }
 }
 list(hessian=(d+t(d))/2,error=err)
}
