########################################################################
#
#	To analyse temperature maxima from Oxford and Worthing.
#	Requires the routines in ev_bivariate.r; also the
#	"ismev" library (available from CRAN)
#
########################################################################
require(ismev)
if (!exists("gev.logdens")) source("ev_bivariate.r")
#
# Plot data, and concatenate into a single vector with a 
# corresponding design matrix
#
t <- 1901:1980
temp.data <- read.table("maxtemps.dat",header=TRUE)
oxford.data <- temp.data$Oxford; worthing.data <- temp.data$Worthing
X <- matrix(rep(c(1,-1),each=80),ncol=1)
y <- c(oxford.data,worthing.data)
par(mfrow=c(2,1))
plot(t,oxford.data,xlab="Year",ylab="Degrees F",type="l",lwd=2,
     ylim=c(75,100))
legend(1950,100,legend=c("Oxford","Worthing"),col=c("black","grey"),
	lty=c(1,2),lwd=2)
lines(t,worthing.data,lwd=2,col="grey",lty=2)
title("Annual maximum temperatures at Oxford and Worthing, 1901-1980")
plot(oxford.data,worthing.data,pch=15,xlab="Oxford (degrees F)",
     ylab="Worthing (degrees F)")
plotfl <- "maxtemps"
dev.copy(pdf,paste(plotfl,".pdf",sep=""),width=11,height=8)
dev.off()
dev.copy(postscript,paste(plotfl,".ps",sep=""),paper="a4",horizontal=TRUE)
dev.off()
#
# Fit a GEV assuming independence, and check
#
model1 <- gev.fit(y,X,mul=1,sigl=1,shl=1)
gev.diag(model1)
plotfl <- "model1_diag"
dev.copy(pdf,paste(plotfl,".pdf",sep=""),width=11,height=8)
dev.off()
dev.copy(postscript,paste(plotfl,".ps",sep=""),paper="a4",horizontal=TRUE)
dev.off()
#
# And a bivariate model. Check the marginal structure using gev.diag() -
# to do this, make an object with all the right bits in it (copy model1,
# and replace the parameter estimates)
#
y1 <- oxford.data
y2 <- worthing.data
X1 <- as.matrix(X[1:80,])
X2 <- as.matrix(X[81:160,])
model2 <- gev2.fit(c(model1$mle,1),y1,y2,X1,X2,mul=1,sigl=1,shl=1)
mod2.check <- model1
mod2.check$mle <- model2$mle
mod2.check$nllh <- model2$nllh
mod2.check$conv <- model2$conv
mod2.check$se <- model2$se
mod2.check$cov <- model2$cov
gev.diag(mod2.check)
plotfl <- "model2_diag"
dev.copy(pdf,paste(plotfl,".pdf",sep=""),width=11,height=8)
dev.off()
dev.copy(postscript,paste(plotfl,".ps",sep=""),paper="a4",horizontal=TRUE)
dev.off()
#
# Print tables of estimates and standard errors
#
estimate <- data.frame(rbind(c(model1$mle,NA),model2$mle))
rownames(estimate) <- c("Independence","Bivariate")
names(estimate) <- c("mu0","mu1","sig0","sig1","xi0","xi1","alpha")
sink("temp_anal.res")
cat("MAXIMUM LIKELIHOOD ESTIMATES:\n")
cat("=============================\n\n")
print(round(estimate,4))

naive <- model1$cov
if (!exists("robust")) {
 robust <- gev.sandwich(model1$mle,y,X,
                       mul=1,sigl=1,shl=1,clusters=rep(t,2))
}
bivariate <- model2$cov
se <- sqrt(rbind(c(diag(naive),NA),c(diag(robust$cov),NA),diag(bivariate)))
se <- data.frame(se)
rownames(se) <- c("Naive","Robust","Bivariate")
names(se) <- names(estimate)
cat("\nSTANDARD ERRORS:\n")
cat("================\n\n")
print(round(se,4))
sink()
#
# NB the "Robust" standard error for xi0 is, unexpectedly, smaller 
# than the "Naive". This has been investigated by computing the 
# sandwich variance estimator assuming that all observations are
# independent (i.e. taking clusters of size 1 - use clusters=1:160
# in gev,sandwich()). Turns out that if you do this, you get a 
# standard error of 0.0379. The robust estimate accounting for
# dependence is larger than this; the difference between "robust"
# and "model-based" estimates suggests a possible problem with 
# the GEV marginal model
# 

#
# OK: now compute the "Cholesky" and "rotation-free" transformation
# matrices. NB the independence Hessian is stored in robust$H;
# the adjusted Hessian is the inverse of robust$cov
#
C.chol <- adj.trans(robust$H,solve(robust$cov),method="cholesky")
C.dil <- adj.trans(robust$H,solve(robust$cov),method="dilation")

#
# And produce some "pilot" 2-D profiles to identify a skewed and an
# elliptical one
#
#par(mfrow=c(2,3))
#proftest <- gev.prof2d(model1,c(1,3),y=y,X=X,mul=1,sigl=1,shl=1,npts=c(5,5))
#contour(proftest$grid1,proftest$grid2,proftest$Logl,
#	xlab=expression(mu[0]),ylab=expression(sigma[0]))
#proftest <- gev.prof2d(model1,c(3,5),y=y,X=X,mul=1,sigl=1,shl=1,npts=c(5,5))
#contour(proftest$grid1,proftest$grid2,proftest$Logl,
#	xlab=expression(sigma[0]),ylab=expression(xi[0]))
#proftest <- gev.prof2d(model1,c(1,5),y=y,X=X,mul=1,sigl=1,shl=1,npts=c(5,5))
#contour(proftest$grid1,proftest$grid2,proftest$Logl,
#	xlab=expression(mu[0]),ylab=expression(xi[0]))
#proftest <- gev.prof2d(model1,c(1,2),y=y,X=X,mul=1,sigl=1,shl=1,npts=c(5,5))
#contour(proftest$grid1,proftest$grid2,proftest$Logl,
#	xlab=expression(mu[0]),ylab=expression(mu[1]))
#proftest <- gev.prof2d(model1,c(3,4),y=y,X=X,mul=1,sigl=1,shl=1,npts=c(5,5))
#contour(proftest$grid1,proftest$grid2,proftest$Logl,
#	xlab=expression(sigma[0]),ylab=expression(sigma[1]))
#proftest <- gev.prof2d(model1,c(5,6),y=y,X=X,mul=1,sigl=1,shl=1,npts=c(5,5))
#contour(proftest$grid1,proftest$grid2,proftest$Logl,
#	xlab=expression(xi[0]),ylab=expression(xi[1]))
#plotfl <- "pilots"
#dev.copy(pdf,paste(plotfl,".pdf",sep=""),width=11,height=8)
#dev.off()
#dev.copy(postscript,paste(plotfl,".ps",sep=""),paper="a4",horizontal=TRUE)
#dev.off()

#
# Right - let's do some serious comparisons of profiles. First
# mu[0] and mu[1]
#
opar <- par(no.readonly=TRUE)
par(omd=c(0.2,0,0.1,0),mfrow=c(1,3),cex=1,mgp=c(2,1,0),mar=c(5,3,2,2))
npts <- c(50,50)
pars <- c(1,2); lower <- c(79,1.4); upper <- c(83,4)
ylim <- c(lower[2],upper[2] + (0.2*(upper[2]-lower[2])))
proftest <- gev.prof2d(model1,pars,y=y,X=X,mul=1,sigl=1,shl=1,
			lower=lower,upper=upper,npts=npts)
contour(proftest$grid1,proftest$grid2,proftest$Logl,
	xlab=expression(mu[0]),ylab=expression(mu[1]),
	levels=-model1$nllh-qchisq(c(0.95,0.99),1),drawlabels=FALSE,
        ylim=ylim,cex.lab=1.2)
points(model1$mle[pars[1]],model1$mle[pars[2]],pch=15)
title(expression(paste("95% & 99% regions for  ",mu[0],"  and  ",mu[1])))
legend(lower[1],ylim[2],legend=c("Naive","Adjusted (Cholesky)",
			"Adjusted (rotation-free)","Bivariate model"),
	lwd=c(1,2,2,2),lty=c(1,2,4,1),
        col=c("black","red","blue","black"),text.width=1.9)
text(model1$mle[pars[1]],model1$mle[pars[2]],
     expression(paste("  ",hat(theta)[IND])),adj=c(0,0))
#
# Cholesky adjustment
#
proftest <- gev.prof2d(model1,pars,y=y,X=X,mul=1,sigl=1,shl=1,
			lower=lower,upper=upper,npts=npts,
			dep.adj=TRUE,C=C.chol)
contour(proftest$grid1,proftest$grid2,proftest$Logl,
	xlab=expression(mu[0]),ylab=expression(xi[0]),
	levels=-model1$nllh-qchisq(c(0.95,0.99),1),drawlabels=FALSE,
	add=TRUE,lwd=2,lty=2,col="red")
#
# Rotation-free adjustment
#
proftest <- gev.prof2d(model1,pars,y=y,X=X,mul=1,sigl=1,shl=1,
			lower=lower,upper=upper,npts=npts,
			dep.adj=TRUE,C=C.dil)
contour(proftest$grid1,proftest$grid2,proftest$Logl,
	xlab=expression(mu[0]),ylab=expression(xi[0]),
	levels=-model1$nllh-qchisq(c(0.95,0.99),1),drawlabels=FALSE,
	add=TRUE,lwd=2,lty=4,col="blue")
#
# And finally the profile from the bivariate model
#
proftest <- gev2.prof2d(model2,pars,y1=y1,y2=y2,X1=X1,X2=X2,mul=1,
			sigl=1,shl=1,lower=lower,upper=upper,npts=npts)
contour(proftest$grid1,proftest$grid2,proftest$Logl,
	xlab=expression(mu[0]),ylab=expression(xi[0]),
	levels=-model2$nllh-qchisq(c(0.95,0.99),1),drawlabels=FALSE,
	add=TRUE,lwd=2)
points(model2$mle[pars[1]],model2$mle[pars[2]],pch=1,mkh=2)
text(model2$mle[pars[1]],model2$mle[pars[2]],expression(hat(theta)[BIV]),
    adj=c(1,1))

#
# Now sigma[0] and sigma[1]
#
pars <- c(3,4); lower <- c(2.9,-0.4); upper <- c(5,1.6)
proftest <- gev.prof2d(model1,pars,y=y,X=X,mul=1,sigl=1,shl=1,
			lower=lower,upper=upper,npts=npts)
contour(proftest$grid1,proftest$grid2,proftest$Logl,
	xlab=expression(sigma[0]),ylab=expression(sigma[1]),
	levels=-model1$nllh-qchisq(c(0.95,0.99),1),drawlabels=FALSE,cex.lab=1.2)
points(model1$mle[pars[1]],model1$mle[pars[2]],pch=15)
text(model1$mle[pars[1]],model1$mle[pars[2]],
     expression(paste("  ",hat(theta)[IND])),adj=c(0,0))
title(expression(paste("95% & 99% regions for  ",sigma[0],"  and  ",sigma[1])))
#
# Cholesky adjustment
#
proftest <- gev.prof2d(model1,pars,y=y,X=X,mul=1,sigl=1,shl=1,
			lower=lower,upper=upper,npts=npts,
			dep.adj=TRUE,C=C.chol)
contour(proftest$grid1,proftest$grid2,proftest$Logl,
	xlab=expression(mu[0]),ylab=expression(xi[0]),
	levels=-model1$nllh-qchisq(c(0.95,0.99),1),drawlabels=FALSE,
	add=TRUE,lwd=2,lty=2,col="red")
#
# Rotation-free adjustment
#
proftest <- gev.prof2d(model1,pars,y=y,X=X,mul=1,sigl=1,shl=1,
			lower=lower,upper=upper,npts=npts,
			dep.adj=TRUE,C=C.dil)
contour(proftest$grid1,proftest$grid2,proftest$Logl,
	xlab=expression(mu[0]),ylab=expression(xi[0]),
	levels=-model1$nllh-qchisq(c(0.95,0.99),1),drawlabels=FALSE,
	add=TRUE,lwd=2,lty=4,col="blue")
#
# And the profile from the bivariate model
#
proftest <- gev2.prof2d(model2,pars,y1=y1,y2=y2,X1=X1,X2=X2,mul=1,
			sigl=1,shl=1,lower=lower,upper=upper,npts=npts)
contour(proftest$grid1,proftest$grid2,proftest$Logl,
	xlab=expression(mu[0]),ylab=expression(xi[0]),
	levels=-model2$nllh-qchisq(c(0.95,0.99),1),drawlabels=FALSE,
	add=TRUE,lwd=2)
points(model2$mle[pars[1]],model2$mle[pars[2]],pch=1,mkh=2)
text(model2$mle[pars[1]],model2$mle[pars[2]],
     expression(hat(theta)[BIV]),adj=c(1,1))

#
# And finally sigma[0] and xi[0]
#
pars <- c(3,5); lower <- c(2.9,-0.38); upper <- c(5,0.08)
proftest <- gev.prof2d(model1,pars,y=y,X=X,mul=1,sigl=1,shl=1,
			lower=lower,upper=upper,npts=npts)
contour(proftest$grid1,proftest$grid2,proftest$Logl,
	xlab=expression(sigma[0]),ylab=expression(xi[0]),
	levels=-model1$nllh-qchisq(c(0.95,0.99),1),drawlabels=FALSE,cex.lab=1.2)
points(model1$mle[3],model1$mle[5],pch=15)
text(model1$mle[pars[1]],model1$mle[pars[2]],expression(hat(theta)[IND]),
    adj=c(1,1))
title(expression(paste("95% & 99% regions for  ",sigma[0],"  and  ",xi[0])))
#
# Cholesky adjustment
#
proftest <- gev.prof2d(model1,pars,y=y,X=X,mul=1,sigl=1,shl=1,
			lower=lower,upper=upper,npts=npts,
			dep.adj=TRUE,C=C.chol)
contour(proftest$grid1,proftest$grid2,proftest$Logl,
	xlab=expression(mu[0]),ylab=expression(xi[0]),
	levels=-model1$nllh-qchisq(c(0.95,0.99),1),drawlabels=FALSE,
	add=TRUE,lwd=2,lty=2,col="red")
#
# Rotation-free adjustment
#
proftest <- gev.prof2d(model1,pars,y=y,X=X,mul=1,sigl=1,shl=1,
			lower=lower,upper=upper,npts=npts,
			dep.adj=TRUE,C=C.dil)
contour(proftest$grid1,proftest$grid2,proftest$Logl,
	xlab=expression(mu[0]),ylab=expression(xi[0]),
	levels=-model1$nllh-qchisq(c(0.95,0.99),1),drawlabels=FALSE,
	add=TRUE,lwd=2,lty=4,col="blue")
#
# And finally the profile from the bivariate model
#
proftest <- gev2.prof2d(model2,pars,y1=y1,y2=y2,X1=X1,X2=X2,mul=1,
			sigl=1,shl=1,lower=lower,upper=upper,npts=npts)
contour(proftest$grid1,proftest$grid2,proftest$Logl,
	xlab=expression(mu[0]),ylab=expression(xi[0]),
	levels=-model2$nllh-qchisq(c(0.95,0.99),1),drawlabels=FALSE,
	add=TRUE,lwd=2)
points(model2$mle[pars[1]],model2$mle[pars[2]],pch=1,mkh=2)
text(model2$mle[pars[1]],model2$mle[pars[2]],
     expression(paste("  ",hat(theta)[BIV])),adj=c(0,0))

par(opar)

plotfl <- "profiles9599"
dev.copy(pdf,paste(plotfl,".pdf",sep=""),width=11,height=8)
dev.off()
dev.copy(postscript,paste(plotfl,".ps",sep=""),paper="a4",horizontal=TRUE)
dev.off()
